"""
This module has functions to process standard table from raw dataframe
containing standard samples
"""
from builtins import object
import json
import numpy as np
import pandas as pd
import warnings as wrngs
from algowrap import err_constants as err_const

from quant import constants


class TableProcessing(object):
    """
    This class has functions to extract, process and format the
    standard table starting from raw dataframe containing standard
    samples.

    Attributes:
        std_df: raw standard dataframe
        conc_col: name of the column containing standard concentrations
    """

    def __init__(self, std_df):
        """
        init function

        Args:
            std_df: raw standard dataframe
        """
        self.std_df = std_df
        column_names = self.std_df.columns.values
        indices = [i for i, s in enumerate(column_names) if
                   constants.MQ_CONCENTRATION in s]
        if len(indices) == 1:
            self.conc_col = column_names[indices[0]]
        else:
            raise KeyError(err_const.INPUTERR8_REPEATED_COLUMN.format(constants.MQ_CONCENTRATION))

    def get_standard_table(self, metab, background_corr):
        """
        Creates standard table for the given base fragment (unlabeled fragment)


        Args:
            metab: unlabeled fragment name
            background_corr: boolean values corresponding to background correction
        Returns:
            std_table: standard table of conc vs intensity
        """
        metab_table = self.std_df[self.std_df[constants.PARENT] == metab]
        intensity_sum = self._compute_sum_by_fragments(metab_table)
        std_table_raw = self._convert_to_std_table(intensity_sum)
        if background_corr:
            std_table_raw = self._intensity_correction_at_origin(std_table_raw)
        std_table_err = self._compute_std_dev_by_cohort(std_table_raw)
        std_table_av = self._compute_average_by_cohort(std_table_raw)
        # converting indices to column
        std_table_av.reset_index(inplace=True)
        std_table = std_table_err.merge(std_table_av)
        return std_table

    def get_all_std_tables_df(self, background_corr):
        """
        creates tables for all unlabeled fragments in the raw dataframe
        and stores in one dataframe

        Args:
            background_corr: boolean values corresponding to background_corr

        Returns:
           ret_df: output dataframe with all tables
           unit: it is the unit of concentration
        """
        list_std_table_df = []
        metab_list = self.std_df[constants.PARENT].unique()
        for metab in metab_list:
            list_std_table_df.append(self.get_standard_table(metab, background_corr))
        ret_df = pd.concat(list_std_table_df)
        # recreate index after the concat
        len_df = len(ret_df)
        idx_list = np.linspace(0, len_df - 1, len_df)
        ret_df[constants.MQ_INDEX] = idx_list
        ret_df.set_index(constants.MQ_INDEX, drop=True, inplace=True)
        # set presicion
        columns = [self.conc_col, constants.AVG_AREA, constants.STD_DEV]
        for column in ret_df.columns.values:
            if constants.MQ_INTENSITY in column:
                columns.append(column)
        ret_df[columns] = ret_df[columns].applymap \
            (lambda x: np.round(x, decimals=constants.PRESICION))
        # find unit of concentration column and send it forward
        unit = ret_df[constants.UNIT].unique()
        unit_check = all(x == unit[0] for x in unit)
        if unit_check:
            unit = unit[0]
        else:
            raise KeyError(err_const.INPUTERR9_DIFFERENT_CONCENTRATION_UNITS)
        # TODO: See if adding tables to a dictionary works better
        ret_df = ret_df.fillna(-1)
        return ret_df, unit

    def _compute_sum_by_fragments(self, metab_table):
        """
        Add intensities for all fragments corresponding to one
        unlabeled fragment. This is done for each replicate.

        Args:
            metab_table: standard dataframe corresponding to one unlabeled
                        fragment

        Returns:
            metab_table_subset: metabolite table with added intensities of all
                                fragments
        """
        intensity_sum = metab_table.groupby([constants.MQ_SAMPLE_NAME])[
            constants.MQ_INTENSITY].sum().reset_index()
        metab_table_subset = metab_table[
            [self.conc_col, constants.UNIT, constants.PARENT,
             constants.MQ_SAMPLE_NAME]]
        metab_table_subset = metab_table_subset.merge(intensity_sum,
                                                      how="inner",
                                                      left_on=constants.MQ_SAMPLE_NAME,
                                                      right_on=constants.MQ_SAMPLE_NAME)
        metab_table_subset = metab_table_subset.drop_duplicates()
        return metab_table_subset

    def _convert_to_std_table(self, sum_metab_table):
        """
        Change the format of table from long format to wide format

        Args:
            sum_metab_table: table for one unlabeled fragment with summed up
            intensities corresponding to its fragments
        Returns:
            std_table: table with different columns for conc and intensities
            correspondng to different replicates
        """
        area_cohort_table = sum_metab_table[
            [constants.PARENT, constants.MQ_INTENSITY, self.conc_col,
             constants.UNIT]]
        area_cohort_table['idx'] = area_cohort_table.groupby(['Concentration', 'Unlabeled Fragment']).cumcount()
        area_cohort_table['idx'] = 'Area_' + area_cohort_table.idx.astype(str)
        unlabeled_fragment = area_cohort_table['Unlabeled Fragment'].unique()
        unit_col = area_cohort_table['Unit'].unique()
        cohort_table = area_cohort_table.pivot(index='Concentration' , columns='idx')[['Area']]
        cohort_table.columns = cohort_table.columns.droplevel(0)
        cohort_table = cohort_table.reset_index()
        cohort_table['Unlabeled Fragment'] = unlabeled_fragment[0]
        cohort_table['Unit'] = unit_col[0]

        return cohort_table

    def _compute_average_by_cohort(self, std_table):
        """
        Create a column containing intensities calculated by taking
        average of intensities for different replicates

        Args:
            std_table: wide format table for one unlabeled fragment with columns
            containg conc and intensities for different replicates

        Returns:
            new_std_table: the original table with new average intensity column
        """
        new_std_table = std_table.copy()
        new_std_table[constants.AVG_AREA] = np.around(
            new_std_table[new_std_table.columns.difference(['Concentration'])].mean(axis=1), constants.PRESICION_TABLE)

        return new_std_table

    def _compute_std_dev_by_cohort(self, std_table):
        """
        Create a column containing standard deviation calculated from
        intensities of different replicates

        Args:
            std_table: wide format table for one unlabeled fragment with columns
            containg conc and intensities for different replicates

        Returns:
            new_std_table: the original table with new std dev column
        """
        new_std_table = std_table.copy()
        new_std_table[constants.STD_DEV] = np.around(
            new_std_table[new_std_table.columns.difference(['Concentration'])].std(axis=1),
                                                     constants.PRESICION_TABLE)
        return new_std_table

    def _intensity_correction_at_origin(self, std_table):
        """
        Correct the intensity column such that at origin the value
        of intensity is zero

        Args:
            std_table: wide format table for one unlabeled fragment with columns
            containg conc and intensities for different replicates

        Returns:
            new_std_table: the original table with corrected intensities
        """
        replicates = list(std_table.columns.values)
        replicates.remove('Unlabeled Fragment')
        replicates.remove('Unit')

        try:
            origin_conc = std_table.iloc[0]
            for col_name in replicates:
                std_table[col_name] = std_table[col_name] - origin_conc[
                    col_name]
            return std_table
        except KeyError:
            wrngs(
                "As 0 concentration not available, proceeding without zero "
                "correction", UserWarning)
            return std_table


def check_all_same_values(values):
    """
    Check all values in list are same and return bool response

    Args:
         list

    return:
        bool
    """
    return all(x == values[0] for x in values)

def assert_curvefit_input_files(input_local_files):
    """
    Verify input files for curvefit. Asserts and raises an exception if the
    required input files are not present.

    Parameters
    ----------
    input_local_files: List
        List of input files with path (from algowrap)
    """

    input_file_names = [constants.MERGED_STD, constants.BIOSAMPLE_DATA]
    if isinstance(input_file_names, list):
        for key in input_file_names:
            if key not in input_local_files:
                raise AssertionError("Input to the application is invalid")
    elif input_file_names not in input_local_files:
            raise AssertionError("Input to the application is invalid")
    else:
        pass


def get_input_dataframes(input_local_files):
    """
    This function reads biosample_data and merged_std json file and convert
    them inoto dataframe

    Args:
        input_local_files : Biosample and merged_std json file

    Returns:
        merged_df: standard sample dataframe with corresponding
                    concentration values
        biosample: biosample dataframe
    """
    assert_curvefit_input_files(input_local_files)
    biosamples = pd.DataFrame(
        json.load(open(input_local_files.get(constants.BIOSAMPLE_DATA))))
    merged_std_df = pd.DataFrame(
        json.load(open(input_local_files.get(constants.MERGED_STD))))
    return biosamples, merged_std_df
